#ifndef __INTEGER_HPP_
#define __INTEGER_HPP_

#include "rings.hpp"

namespace Rings {

/**
  @brief encapsulates integers under the \c Ring_Element rubric
  @details The template is used to specify the base numeric type: \c int,
      \c uint32_t, etc. In theory you could do something stupid like instantiate
      an \c Integer<double> but why? Anyway, this class may be extended
      to have methods that would rule that out; e.g., \c T could require the
      \c % operator and good luck doing that with your silly little \c double.
*/
template<typename T>
class Integer : virtual public Integral_Domain_Element {

private:

  T value;

public:

  /** @brief initializes to zero */
  Integer<T>();
  /** @brief initializes to the given value */
  Integer<T>(T);
  /** @brief copy constructor */
  Integer<T>(const Integer<T> & other);

  /** @brief assignment operator may be needed */
  const Integer<T> & operator = (const Ring_Element &);

  virtual bool is_one() const override;
  virtual bool is_zero() const override;
  virtual bool has_inverse() const override;

  virtual bool operator == (const Ring_Element &) const override;
  virtual bool operator != (const Ring_Element &) const override;

  virtual Ring_Element & operator + (const Ring_Element &) const override;
  virtual Ring_Element & operator - (const Ring_Element &) const override;
  virtual Ring_Element & operator * (const Ring_Element &) const override;

  /** @brief returns the value of \c this */
  T get_value() const;

};

template<typename T>
Integer<T>::Integer() : value(0) { }

template<typename T>
Integer<T>::Integer(T v) : value(v) { }

template<typename T>
Integer<T>::Integer(const Integer<T> & other) : value(other.value) { }

template<typename T>
bool Integer<T>::is_one() const { return value == 1; }

template<typename T>            
bool Integer<T>::is_zero() const { return value == 0; }

template<typename T>
bool Integer<T>::has_inverse() const { return value == 1 or value == -1; }

template<typename T>
Ring_Element & Integer<T>::operator + (const Ring_Element & other) const {
  static Integer<T> result;
  auto o = dynamic_cast<const Integer<T> &>(other);
  result.value = value + o.value;
  return result;
}

template<typename T>
Ring_Element & Integer<T>::operator - (const Ring_Element & other) const {
  static Integer<T> result;
  auto o = dynamic_cast<const Integer<T> &>(other);
  result.value = value - o.value;
  return result;
}

template<typename T>
Ring_Element & Integer<T>::operator * (const Ring_Element & other) const {
  static Integer<T> result;
  auto & o = dynamic_cast<const Integer<T> &>(other);
  result.value = value * o.value;
  return result;
}

template<typename T>
const Integer<T> & Integer<T>::operator = (const Ring_Element & other) {
  if (*this != other) {
    auto & o = dynamic_cast<const Integer<T> &>(other);
    value = o.value;
  }
  return *this;
}

template<typename T>
bool Integer<T>::operator == (const Ring_Element & other) const {
  auto & o = dynamic_cast<const Integer<T> &>(other);
  return value == o.value;
}

template<typename T>
bool Integer<T>::operator != (const Ring_Element & other) const {
  auto & o = dynamic_cast<const Integer<T> &>(other);
  return value != o.value;
}

template<typename T>
T Integer<T>::get_value() const { return value; }

/**
  @relates Integer
  @brief prints the value of the given integer
*/
template <typename T>
ostream & operator << (ostream & os, Integer<T> i) {
  os << i.get_value();
  return os;
}

}

#endif