#ifndef __MOD_
#define __MOD_

#include <stdexcept>
using std::domain_error;

#include <iostream>
using std::ostream;

#include "../gcd/gcd_template.hpp"

#include "rings.hpp"

namespace Rings {

/**
  @brief templated modular arithmetic, making no assumption about the modulus
  @details The first template value indicates the base type (e.g., \c int,
      \c long). This code assumes you&rsquo;re working with something \c int
      -like so no guarantees if you try something else. The second template
      indicates the modulus. See the warning.
  @warning If the modulus is nonzero then you get what you&rsquo;re asking for!
*/
template<typename T, T m>
class Mod : virtual public Commutative_Ring_Element {
protected:

  /** @brief value of the element */
  T value;
  /** @brief whether the element is invertible */
  bool invertible;

  /** @brief ensures value is at least 0 but less than modulus */
  void adjust_value() {
    value %= m;
    if (value < 0)
      value += m;
  }
  /**
    @brief decides whether \c this is invertible and sets flag accordingly;
    \c True iff \f$\gcd(v,m)=1\f$ (where \f$v\f$ is value)
  */
  virtual void check_inverse() { invertible = 1 == gcd<T>(value, m); }
  /**
    @brief checks modulus to ensure we aren't trying to do the impossible
    @details To minimize the penalty of checking and throwing,
        we use this only with the default constructor and initialize one below
  */
  void check_modulus() {
    if (m == 0) throw domain_error("Attempted to create a modulus of 0");
  }

public:

  /** @brief initializes to 0 */
  Mod<T,m>();
  /** @brief initializes to give value */
  Mod<T, m>(T v);
  /** @brief copy constructor */
  Mod<T, m>(const Mod<T, m> & other);

  /** @brief \c True iff assigned value is congruent to 1 */
  virtual bool is_one() const override;
  /** @brief \c True iff assigned value is congruent to 0 */
  virtual bool is_zero() const override;
  /** @brief \c True iff assigned value is relatively prime to modulus */
  virtual bool has_inverse() const override;

  /** @brief implements modular addition */
  virtual Mod<T, m> & operator + (const Ring_Element & other) const override;
  /** @brief implements modular subtraction */
  virtual Mod<T, m> & operator - (const Ring_Element & other) const override;
  /** @brief implements modular multiplication */
  virtual Mod<T, m> & operator * (const Ring_Element & other) const override;
  /** @brief multiplication of value with element of base type */
  virtual Mod<T, m> & operator * (const T & other) const;

  /** @brief assignment operator may be needed */
  const Mod<T,m> & operator = (const Ring_Element &);

  virtual bool operator == (const Ring_Element &) const override;
  virtual bool operator != (const Ring_Element &) const override;

  /** @brief returns value of this modulus */
  T get_value() const;

};

template<typename T, T m>
Mod<T,m>::Mod() : value(0) { check_modulus(); check_inverse(); }

template<typename T, T m>
Mod<T,m>::Mod(T v) : value(v) { adjust_value(); check_inverse(); }

template<typename T, T m>
Mod<T,m>::Mod(const Mod<T, m> & other)
    : value(other.value), invertible(other.invertible)
{ }

template<typename T, T m>
bool Mod<T,m>::is_one() const { return value == 1; }

template<typename T, T m>
bool Mod<T,m>::is_zero() const { return value == 0; }

template<typename T, T m>
bool Mod<T,m>::has_inverse() const { return invertible; }

template<typename T, T m>
Mod<T, m> & Mod<T,m>::operator + (const Ring_Element & other) const
{
  static Mod<T, m> result;
  auto o = dynamic_cast<const Mod<T, m> &>(other);
  result.value = value + o.value;
  result.adjust_value();
  return result;
}

template<typename T, T m>
Mod<T, m> & Mod<T,m>::operator - (const Ring_Element & other) const
{
  static Mod<T, m> result;
  auto o = dynamic_cast<const Mod<T, m> &>(other);
  result.value = value - o.value;
  result.adjust_value();
  return result;
}

template<typename T, T m>
Mod<T, m> & Mod<T,m>::operator * (const Ring_Element & other) const
{
  static Mod<T, m> result;
  auto o = dynamic_cast<const Mod<T, m> &>(other);
  result.value = value * o.value;
  result.adjust_value();
  return result;
}

template<typename T, T m>
Mod<T, m> & Mod<T,m>::operator * (const T & other) const {
  static Mod<T, m> result;
  result.value = value * other;
  result.adjust_value();
  return result;
}

template<typename T, T m>
const Mod<T,m> & Mod<T,m>::operator = (const Ring_Element & other) {
  if (*this != other) {
    const Mod<T,m> & o = dynamic_cast<const Mod<T,m> & >(other);
    value = o.value;
    invertible = o.invertible;
  }
  return *this;
}

template<typename T, T m>
bool Mod<T,m>::operator == (const Ring_Element & other) const {
  const Mod<T,m> & o = dynamic_cast<const Mod<T,m> & >(other);
  return value == o.value;
}

template<typename T, T m>
bool Mod<T,m>::operator != (const Ring_Element & other) const {
  const Mod<T,m> & o = dynamic_cast<const Mod<T,m> & >(other);
  return value != o.value;
}

template<typename T, T m>
T Mod<T,m>::get_value() const { return value; }

/**
  @relates Mod
  @brief prints only value, not the modulus
*/
template <typename T, T m>
ostream & operator << (ostream & os, const Mod<T, m> & x) {
  os << x.get_value();
  return os;
}

/**
  @relates Mod
  @brief used to prevent the use of a zero modulus
*/
template<typename T, T m>
constexpr Mod<T,m> prevent_zero_modulus;

}

#endif