#ifndef __MODP_HPP_
#define __MODP_HPP_

#include <stdexcept>
using std::domain_error;

#include <vector>
using std::vector;

#include "mod.hpp"

namespace Rings {

/**
  @brief templated modular arithmetic, modulo a prime \f$p\f$
  @details The first template value indicates the base type (e.g., \c int,
      \c long). This code assumes you&rsquo;re working with something \c int
      -like so no guarantees if you try something else. The second template
      indicates the modulus, which should be a prime number. This guarantees
      inverses for every nonzero element of the field, but see the warning.
  @warning If the modulus is nonzero then you get what you&rsquo;re asking for!
*/
template<typename T, T p>
class Modp : virtual public Field_Element, virtual public Mod<T, p> {

protected:

  using Mod<T,p>::value;
  using Mod<T,p>::invertible;

  /** @brief lists inverses for all nonzero elements */
  static vector<T> inverses;

  /** @brief redefines \c Mod&rsquo;s \c check_inverse() to do nothing */
  virtual void check_inverse() noexcept override { }
  /** @brief checks whether inverse are set up, and if not sets them up */
  void check_inverses();

public:

  using Field_Element::has_inverse;

  /** @brief initializes to zero */
  Modp<T,p>();
  /** @brief initializes to given value */
  Modp<T,p>(T);
  /**
    @brief copy constructor of parent type (should be safe, thanks to modulus)
  */
  Modp<T,p>(const Mod<T,p> & other);
  /** @brief copy constructor of same type*/
  Modp<T,p>(const Modp<T,p> & other);

  virtual Modp<T,p> & operator * (const T & other) const override;
  virtual Modp<T,p> & operator / (const Field_Element & other) const override;
  virtual Modp<T,p> & inverse() const override;

  virtual bool has_inverse() const override;

};

template <typename T, T p>
void Modp<T,p>::check_inverses() {
  if (inverses.size() == 0) {
    inverses.resize(p);
    inverses[1] = 1;
    inverses[p - 1] = p - 1;
    for (T i = 2; i < p - 1; ++i) {
      T a, b;
      T d = gcd<T>(i, p, a, b);
      if (d != 1)
          throw domain_error("Attempted a non-prime modulus for Modp");
      if (a < 0) a += p;
      inverses[i] = a;
    }
  }
}

template <typename T, T p>
Modp<T,p>::Modp() : Mod<T,p>() { invertible = true; check_inverses(); }

template <typename T, T p>
Modp<T,p>::Modp(T v) : Mod<T,p>(v) {
  invertible = true;
  check_inverses();
}

template <typename T, T p>
Modp<T,p>::Modp(const Modp<T,p> & other) : Mod<T,p>(other) {
  invertible = true;
  check_inverses();
}

template <typename T, T p>
Modp<T,p>::Modp(const Mod<T,p> & other) : Mod<T,p>(other) {
  invertible = true;
  check_inverses();
}

template <typename T, T p>
Modp<T,p> & Modp<T,p>::operator * (const T & other) const {
  static Modp<T,p> result(value * other);
  return result;
}

template <typename T, T p>
Modp<T,p> & Modp<T,p>::operator / (const Field_Element & other)
    const
{
  static Modp<T,p> result;
  result.value
      = value * inverses[dynamic_cast<const Modp<T,p> &>(other).value];
  result.adjust_value();
  return result;
}

template <typename T, T p>
Modp<T,p> & Modp<T,p>::inverse() const {
  static Modp<T,p> result;
  result.value = inverses[value];
  return result;
}

template <typename T, T p>
bool Modp<T,p>::has_inverse() const { return not is_zero(); }

template<typename T, T p> vector<T> Modp<T,p>::inverses;

/**
  @relates Modp
  @brief used to force an initialization of inverses at compilation
*/
template<typename T, T p>
constexpr Modp<T,p> initialize_inverses;

}

#endif