#ifndef __POLYNOMIAL_H_
#define __POLYNOMIAL_H_

#include <initializer_list>
using std::initializer_list;
#include <vector>
using std::vector;

#include <iostream>
using std::ostream;

#include "rings.hpp"

namespace Rings {

/**
  @brief a dense univariate polynomial, templated according to the
    coefficient&rsquo;s type
  @brief The coefficient should descend from Ring_Element.
*/
template<typename R>
class Polynomial : virtual public Commutative_Ring_Element {

protected:

  DEGREE_TYPE deg; /**< @brief polynomial&rsquo;s degree */
  vector<R> coeffs; /**< @brief polynomial&rsquo;s coefficients */

  /** @brief storage for previously used divisor */
  static Polynomial<R> last_divisor;
  /** @brief storage for previously computed quotient */
  static Polynomial<R> last_quotient;
  /** @brief storage for previously computed remainder */
  static Polynomial<R> last_remainder;

  /** @brief sets all coefficients to zero */
  void clear_coeffs();
  /** @brief sets the degree to the specified value and resizes if necessary */
  void set_degree(DEGREE_TYPE);
  /**
    @brief verifies that the leading coefficient is zero;
    if not, decreases degree until either it is or the degree is zero
    */
  void verify_degree();
  /**
    @brief utility function that obtains
    both divisor and remainder from division
  */
  void common_division(const Polynomial<R> &) const;

public:

  // constructors

  /** @brief generates a zero polynomial */
  Polynomial();
  /** @brief copies the coefficients of the given polynomial to this one */
  Polynomial(const Polynomial<R> &);
  /** @brief initializes coefficients to the ones listed, constant term first */
  Polynomial(const initializer_list<R> &);
  /** @brief initializes coefficients to the ones listed, constant term first */
  Polynomial(const vector<R> &);

  // required methods

  /** @brief indicates the polynomial's degree */
  DEGREE_TYPE degree() const;
  /** @brief returns the coefficient at the specified degree */
  const R & coeff(DEGREE_TYPE) const;
  /** @brief computes the value of the polynomial at the given point */
  const R & operator () (const R &) const;

  /** @brief sets the value of the coefficient at the given degree */
  void set_coeff(DEGREE_TYPE, const R &);

  /** @brief assignment operator may be needed */
  const Polynomial<R> & operator = (const Ring_Element &);

  /**
    @brief returns true if and only if @c this is constant and has the same
      value as the given constant/scalar
  */
  bool operator == (const R &) const;

  /** @brief quotient from dividing @c this by the given divisor */
  const Polynomial<R> & operator / (const Polynomial<R> &) const;

  // overriding Commutative_Ring_Element pure virtual

  virtual bool is_one() const override;
  virtual bool is_cancellable() const override;

  virtual bool operator == (const Ring_Element &) const override;

  virtual Polynomial<R> & operator + (const Ring_Element &) const override;
  virtual Polynomial<R> & operator - (const Ring_Element &) const override;
  virtual Polynomial<R> & operator * (const Ring_Element &) const override;

};

template<typename R>
Polynomial<R>::Polynomial(const Polynomial<R> & other)
    : deg(other.deg), coeffs(deg + 1)
{
  for (unsigned i = 0; i <= deg; ++i)
    coeffs[i] = other.coeffs[i];
}

template<typename R>
Polynomial<R>::Polynomial(const initializer_list<R> & list)
    : deg(list.size() - 1), coeffs(deg + 1)
{
  DEGREE_TYPE i = 0;
  for (auto c : list) {
    coeffs[i] = c;
    ++i;
  }
  verify_degree();
}

template<typename R>
void Polynomial<R>::verify_degree() {
  unsigned i;
  for (i = deg; i > 0 and coeffs[i].is_zero(); --i) { /* do nothing */ }
  deg = i;
}

template<typename R>
void Polynomial<R>::set_degree(DEGREE_TYPE d) {
  if (deg < d)
    coeffs.resize(d + 1);
  deg = d;
}

template<typename R>
void Polynomial<R>::clear_coeffs() {
  R zero;
  for (DEGREE_TYPE i = 0; i <= deg; ++i)
    coeffs[i] = zero;
}

template<typename R>
DEGREE_TYPE Polynomial<R>::degree() const { return deg; }

template<typename R>
const R & Polynomial<R>::coeff(DEGREE_TYPE d) const { return coeffs[d]; }

template<typename R>
const R & Polynomial<R>::operator() (const R & a) const {
  static R result;
  for (DEGREE_TYPE i = 0; i <= deg; ++i) {
    if (not coeffs[i].is_zero()) {
      R term(a);
      for (DEGREE_TYPE j = 0; j <= deg; ++j)
        term = term * a;
      term = term * coeffs[i];
      result = result + term;
    }
  }
  return result;
}

template<typename R>
void Polynomial<R>::set_coeff(DEGREE_TYPE d, const R & value) {
  coeffs[d] = value;
}

template<typename R>
const Polynomial<R> & Polynomial<R>::operator = (const Ring_Element & o) {
  if (*this != o) {
    const Polynomial<R> & other = dynamic_cast<const Polynomial<R> &>(o);
    coeffs.resize(other.degree() + 1);
    deg = other.degree();
    for (DEGREE_TYPE i = 0; i < deg; ++i)
      coeffs[i] = other.coeffs[i];
  }
  return *this;
}

template<typename R>
bool Polynomial<R>::operator == (const Ring_Element & o) const {
  auto & other = dynamic_cast<const Polynomial<R> &>(o);
  bool result = deg == other.deg;
  for (DEGREE_TYPE i = 0; result and i < deg; ++i)
    result = coeffs[i] == other.coeffs[i];
  return result;
}

template<typename R>
bool Polynomial<R>::operator == (const R & value) const {
  return deg == 0 and coeffs[0] == value;
}

template<typename R>
bool Polynomial<R>::is_one() const {
  return degree() == 0 and coeffs[0].is_one();
}

template<typename R>
bool Polynomial<R>::is_cancellable() const {
  bool result = false;
  for (DEGREE_TYPE i = 0; i < deg and not result; ++i)
    result = coeffs[i].is_cancellable();
  return result;
}

template<typename R>
Polynomial<R> & Polynomial<R>::operator + (const Ring_Element & o)
    const
{
  static Polynomial<R> result;
  const Polynomial<R> & other = dynamic_cast<const Polynomial<R> &>(o);
  DEGREE_TYPE d = degree();
  if (d < other.degree()) d = other.degree();
  result.set_degree(d);
  result.clear_coeffs();
  DEGREE_TYPE i;
  for (i = 0; i <= degree() and i <= other.degree(); ++i)
    result.coeffs[i] = coeffs[i] + other.coeffs[i];
  for ( /* */ ; i <= degree(); ++i)
    result.coeffs[i] = coeffs[i];
  for ( /* */ ; i <= other.degree(); ++i)
    result.coeffs[i] = other.coeffs[i];
  result.verify_degree();
  return result;
}

template<typename R>
void Polynomial<R>::common_division(const Polynomial<R> & other) const {
  /* work out on your own and fill in */
}

template<typename R>
const Polynomial<R> & Polynomial<R>::operator / (const Polynomial<R> & other)
    const
{
  static Polynomial<R> result;
  if (other == last_divisor)
    result = last_quotient;
  else {
    common_division(other);
    result = last_quotient;
  }
  return result;
}

/**
  @relates Polynomial
  @brief prints a polynomial in customary fashion
*/
template<typename R>
ostream & operator << (ostream & os, const Polynomial<R> & p) {
  bool first = true;
  for (DEGREE_TYPE d = p.degree(); d > 0; --d) {
    if (not p.coeff(d).is_zero()) {
      if (first) first = false;
      else {
        os << " + ";
      }
      if (not p.coeff(d).is_one()) os << p.coeff(d) << " ";
      os << "x";
      if (d > 1)
        os << "^" << d;
    }
  }
  if (p.degree() == 0 or not p.coeff(0).is_zero()) {
    if (not first) os << " + ";
    os << p.coeff(0);
  }
  return os;
}

}

#endif