#ifndef __RINGS_HPP_
#define __RINGS_HPP_

namespace Rings {

/**
  @brief a class for elements with the capabilities of ring arithmetic
*/
class Ring_Element {
public:
  /** @brief should be \c True iff element is multiplicative identity */
  virtual bool is_one() const = 0;
  /** @brief should be \c True iff element is additive identity */
  virtual bool is_zero() const = 0;
  /** @brief should be \c True iff element has a multiplicative inverse */
  virtual bool has_inverse() const = 0;
  /** @brief should be \c True iff element commutes under multiplication */
  virtual bool is_commutative() const { return false; }
  /**
    @brief should be \c True iff element can cancel across an equation;
        e.g., \f$ax=ay\Rightarrow x=y\f$.
  */
  virtual bool is_cancellable() const { return false; }

  /** @brief comparison: other element has same value */
  virtual bool operator == (const Ring_Element &) const = 0;
  /** @brief comparison: other element has different value */
  virtual bool operator != (const Ring_Element &) const = 0;

  /** @brief addition: other element should be of same type, use a cast */
  virtual const Ring_Element & operator + (const Ring_Element & ) const = 0;
  /** @brief subtraction: other element should be of same type, use a cast */
  virtual const Ring_Element & operator - (const Ring_Element & ) const = 0;
  /** @brief multiplicationL other element should be of same type, use a cast */
  virtual const Ring_Element & operator * (const Ring_Element & ) const = 0;
};

/** @brief elements of this type should have commutative multiplication */
class Commutative_Ring_Element : public Ring_Element {
public:
  /** @brief Duh. */
  virtual bool is_commutative() const override { return true; }
};

/**
  @brief elements of this type should be cancellable;
      there should be no zero divisors
  @details Zero divisors are any \f$a,b\neq0\f$ such that \f$ab=0\f$.
      By &ldquo;cancellable&rdquo; we mean the following:
      \f$ax=ay\Rightarrow a(x-y)=0\f$, so if
      \f$a\neq0\f$ then \f$x-y=0\Rightarrow x=y\f$; that is, we cancelled.
*/
class Integral_Domain_Element : virtual public Commutative_Ring_Element {
public:
  /**
    @brief integral domains are commutative rings without zero divisors,
        so the element should be cancellable (see description of class)
  */
  virtual bool is_cancellable() const override { return not is_zero(); }
};

/** @brief a field is an integral domain whose nonzero elements have inverses */
class Field_Element : virtual public Integral_Domain_Element {

public:

  /** @brief fields are integral domains where nonzero elements have inverses */
  virtual bool has_inverse() const override { return not is_zero(); }

  /** @brief division: other element should be of same type, use a cast */
  virtual Field_Element & operator / (const Field_Element &) const = 0;
  /** @brief multiplicative inverse */
  virtual Field_Element & inverse() const = 0;

};

}

#endif