#include <cstdint>

#include <string>
using std::string;
#include <iostream>
using std::cout; using std::endl;

#include "mod.hpp"
using Rings::Mod;
#include "modp.hpp"
using Rings::Modp;
#include "rational.hpp"
using Rings::Rational;
#include "polynomial.hpp"
using Rings::Polynomial;

/** @brief converts @p b to a more easily-read string */
string bstring(bool b) {
  const static string yes = "true";
  const static string no  = "false";

  if (b) return yes;
  else return no;
}

/** @brief tests the basic construction of polynomials */
void test_basic_construction() {
  // modulo 6 first
  Polynomial<Mod<int32_t, 6> > p = { 2, 4, 0, 2, 4 };
  vector<Mod<int32_t, 6> > v = { 3, 0, 3, 0, 3 };
  Polynomial<Mod<int32_t, 6> > q(v);
  Polynomial<Mod<int32_t, 6> > r(p);
  cout << "first and third should be the same:\n";
  cout << '\t' << p << endl;
  cout << '\t' << q << endl;
  cout << '\t' << r << endl;
  cout << "is " << p << " cancellable? (should be false)\n";
  cout << '\t' << bstring(p.is_cancellable()) << endl;

  // rational next
  // need coefficients first
  Rational<int16_t> a(3,2);
  Rational<int16_t> b(4,7);
  Polynomial<Rational<int16_t> > f = { a, 1, 0, a };
  Polynomial<Rational<int16_t> > g = { 0, 2, b };
  cout << "rational polynomials:\n";
  cout << '\t' << f << endl;
  cout << '\t' << g << endl;
  cout << "is " << f << " cancellable? (should be true)\n";
  cout << '\t' << bstring(f.is_cancellable()) << endl;
}

/** @brief tests the comparison operators */
void test_comparison() {
  // modulo 6 first
  Polynomial<Mod<int32_t, 6> > p = { 2, 4, 0, 2, 4 };
  vector<Mod<int32_t, 6> > v = { 3, 0, 3, 0, 3 };
  Polynomial<Mod<int32_t, 6> > q(v);
  Polynomial<Mod<int32_t, 6> > r(p);
  cout << "second and last should be true, others false\n";
  cout << '\t' << bstring(p == q) << endl;
  cout << '\t' << bstring(p == r) << endl;
  cout << '\t' << bstring(q == r) << endl;

  // rational next
  // need coefficients first
  Rational<int16_t> a(3,2);
  Rational<int16_t> b(4,7);
  Polynomial<Rational<int16_t> > f = { a, 1, 0, a };
  Polynomial<Rational<int16_t> > g = { 0, 2, b };
  cout << '\t' << bstring(f == g) << endl;
  cout << '\t' << bstring(f == f) << endl;
}

/** @brief tests basic arithmetic (not division) */
void test_arithmetic() {
  // modulo 6 first -- funny things will happen
  Polynomial<Mod<int32_t, 6> > p = { 2, 4, 0, 4, 2 };
  vector<Mod<int32_t, 6> > v = { 3, 0, 3, 0, 3 };
  Polynomial<Mod<int32_t, 6> > q(v);
  Polynomial<Mod<int32_t, 6> > r(p);
  cout << "Working modulo 6:\n";
  cout << "\t( " << p << " ) + ( " << q << " ) = " << p + q << endl;
  cout << "\t( " << p << " ) - ( " << q << " ) = " << p - q << endl;
  cout << "\t( " << p << " ) * ( " << q << " ) = " << p *q << endl;

  // rational next
  // need coefficients first
  Rational<int16_t> a(3,2);
  Rational<int16_t> b(4,7);
  Polynomial<Rational<int16_t> > f = { a, 1, 0, a };
  Polynomial<Rational<int16_t> > g = { 0, 2, b };
  cout << "Working over the rationals:\n";
  cout << "\t( " << f << " ) + ( " << f << " ) = " << f + f << endl;
  cout << "\t( " << f << " ) - ( " << f << " ) = " << f - f << endl;
  cout << "\t( " << f << " ) + ( " << g << " ) = " << f + g << endl;
  cout << "\t( " << f << " ) * ( " << g << " ) = " << f * g << endl;
}

/** @brief tests division with remainder */
void test_division() {
  // modulo 5 first -- non-prime could get ugly
  Polynomial<Mod<int32_t, 5> > p = { 2, 4, 0, 4, 2, 0, 1 };
  vector<Mod<int32_t, 5> > v = { 3, 0, 3, 0, 1 };
  Polynomial<Mod<int32_t, 5> > q(v);
  cout << "Working modulo 5:\n";
  cout << "\t( " << p << " ) / ( " << q << " ) = " << p / q << endl;
  cout << "\t( " << p << " ) % ( " << q << " ) = " << p % q << endl;

  // rational next
  // need coefficients first
  Rational<int16_t> a(3,2);
  Rational<int16_t> b(4,7);
  Polynomial<Rational<int16_t> > f = { a, a, 0, 1 };
  Polynomial<Rational<int16_t> > g = { b, 2, 1 };
  cout << "Working over the rationals:\n";
  cout << "\t( " << f << " ) / ( " << f << " ) = " << f / f << endl;
  cout << "\t( " << f << " ) % ( " << f << " ) = " << f % f << endl;
  cout << "\t( " << f << " ) / ( " << g << " ) = " << f / g << endl;
  cout << "\t( " << f << " ) % ( " << g << " ) = " << f % g << endl;
}

int main() {
  test_basic_construction();
  test_comparison();
  test_arithmetic();
  test_division();
}